import pytest

from anoncreds.protocol.types import ProofRequest, PredicateGE, AttributeInfo
from anoncreds.test.conftest import presentProofAndVerify


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testNoPredicates(prover1, verifier, claimsProver1):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={'uuid1': AttributeInfo(
                                    name='name'), 'uuid2': AttributeInfo(name='status')},
                                predicates={})

    assert await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testGePredicate(prover1, verifier, claimsProver1):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('period', 5)})
    assert await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testGePredicateForEqual(prover1, verifier, claimsProver1):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('period', 8)})
    assert await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testGePredicateNegative(prover1, verifier, claimsProver1):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('period', 9)})
    with pytest.raises(ValueError):
        await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testMultipleGePredicate(prover1, verifier, claimsProver1):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('period', 5)})
    await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testMultipleGePredicateMultipleRevealed(prover1, verifier,
                                                  claimsProver1):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={'attr_uuid1': AttributeInfo(name='name'),
                                                      'attr_uuid2': AttributeInfo(name='status')},
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('period', 5)})
    await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testMultipleGePredicateNegative(prover1, verifier, claimsProver1):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('period', 9)})
    with pytest.raises(ValueError):
        await presentProofAndVerify(verifier, proofRequest, prover1)



from abc import abstractmethod

from anoncreds.protocol.types import Attribs, SchemaKey


class AttributeRepo:
    @abstractmethod
    def getAttributes(self, schemaKey: SchemaKey, userId) -> Attribs:
        raise NotImplementedError

    @abstractmethod
    def addAttributes(self, schemaKey: SchemaKey, userId,
                      attributes: Attribs):
        raise NotImplementedError


class AttributeRepoInMemory(AttributeRepo):
    def __init__(self):
        self.attributes = {}

    def getAttributes(self, schemaKey: SchemaKey, userId) -> Attribs:
        return self.attributes.get((schemaKey, userId))

    def addAttributes(self, schemaKey: SchemaKey, userId,
                      attributes: Attribs):
        self.attributes[(schemaKey, userId)] = attributes



from abc import abstractmethod
from typing import Dict, Any

from anoncreds.protocol.exceptions import SchemaNotFoundError
from anoncreds.protocol.types import ID, PublicKey, RevocationPublicKey, \
    Schema, Tails, Accumulator, \
    AccumulatorPublicKey, TimestampType, SchemaKey


class PublicRepo:
    # GET

    @abstractmethod
    async def getSchema(self, schemaId: ID) -> Schema:
        raise NotImplementedError

    @abstractmethod
    async def getPublicKey(self,
                           schemaId: ID,
                           signatureType='CL') -> PublicKey:
        raise NotImplementedError

    @abstractmethod
    async def getPublicKeyRevocation(self,
                                     schemaId: ID,
                                     signatureType='CL') -> RevocationPublicKey:
        raise NotImplementedError

    @abstractmethod
    async def getPublicKeyAccumulator(self,
                                      schemaId: ID) -> AccumulatorPublicKey:
        raise NotImplementedError

    @abstractmethod
    async def getAccumulator(self, schemaId: ID) -> Accumulator:
        raise NotImplementedError

    @abstractmethod
    async def getTails(self, schemaId: ID) -> Tails:
        raise NotImplementedError

    # SUBMIT

    @abstractmethod
    async def submitSchema(self,
                           schema: Schema) -> Schema:
        raise NotImplementedError

    @abstractmethod
    async def submitPublicKeys(self,
                               schemaId: ID,
                               pk: PublicKey,
                               pkR: RevocationPublicKey = None,
                               signatureType='CL') -> (
            PublicKey, RevocationPublicKey):
        raise NotImplementedError

    @abstractmethod
    async def submitAccumulator(self, schemaId: ID,
                                accumPK: AccumulatorPublicKey,
                                accum: Accumulator, tails: Tails) -> \
            AccumulatorPublicKey:
        raise NotImplementedError

    @abstractmethod
    async def submitAccumUpdate(self, schemaId: ID, accum: Accumulator,
                                timestampMs: TimestampType):
        raise NotImplementedError


class PublicRepoInMemory(PublicRepo):
    def __init__(self):
        self._schemasByKey = {}
        self._schemasById = {}
        self._pks = {}
        self._pkRs = {}
        self._accums = {}
        self._accumPks = {}
        self._tails = {}
        self._schemaId = 1
        self._pkId = 1
        self._pkRId = 1
        self._acumPkId = 1

    # GET

    async def getSchema(self, schemaId: ID) -> Schema:
        if schemaId.schemaKey and schemaId.schemaKey in self._schemasByKey:
            return self._schemasByKey[schemaId.schemaKey]

        if schemaId.schemaId and schemaId.schemaId in self._schemasById:
            return self._schemasById[schemaId.schemaId]

        raise SchemaNotFoundError(
            'No schema with ID={} and key={}'.format(
                schemaId.schemaId,
                schemaId.schemaKey))

    async def getPublicKey(self,
                           schemaId: ID,
                           signatureType='CL') -> PublicKey:
        return await self._getValueForId(self._pks, schemaId)

    async def getPublicKeyRevocation(self,
                                     schemaId: ID,
                                     signatureType='CL') -> RevocationPublicKey:
        return await self._getValueForId(self._pkRs, schemaId)

    async def getPublicKeyAccumulator(self,
                                      schemaId: ID) -> AccumulatorPublicKey:
        return await self._getValueForId(self._accumPks, schemaId)

    async def getAccumulator(self, schemaId: ID) -> Accumulator:
        return await self._getValueForId(self._accums, schemaId)

    async def getTails(self, schemaId: ID) -> Tails:
        return await self._getValueForId(self._tails, schemaId)

    # SUBMIT

    async def submitSchema(self,
                           schema: Schema) -> Schema:
        schema = schema._replace(seqId=self._schemaId)
        self._schemaId += 1
        self._schemasByKey[schema.getKey()] = schema
        self._schemasById[schema.seqId] = schema
        return schema

    async def submitPublicKeys(self,
                               schemaId: ID,
                               pk: PublicKey,
                               pkR: RevocationPublicKey = None,
                               signatureType='CL') -> (
            PublicKey, RevocationPublicKey):
        pk = pk._replace(seqId=self._pkId)
        self._pkId += 1
        await self._cacheValueForId(self._pks, schemaId, pk)

        if pkR:
            pkR = pkR._replace(seqId=self._pkRId)
            self._pkRId += 1
            await self._cacheValueForId(self._pkRs, schemaId, pkR)

        return pk, pkR

    async def submitAccumulator(self, schemaId: ID,
                                accumPK: AccumulatorPublicKey,
                                accum: Accumulator,
                                tails: Tails) -> AccumulatorPublicKey:
        accumPK = accumPK._replace(seqId=self._acumPkId)
        self._acumPkId += 1
        await self._cacheValueForId(self._accums, schemaId, accum)
        accumPk = await self._cacheValueForId(self._accumPks, schemaId,
                                              accumPK)
        await self._cacheValueForId(self._tails, schemaId, tails)
        return accumPk

    async def submitAccumUpdate(self, schemaId: ID, accum: Accumulator,
                                timestampMs: TimestampType):
        await self._cacheValueForId(self._accums, schemaId, accum)

    async def _getValueForId(self, dictionary: Dict[SchemaKey, Any],
                             schemaId: ID) -> Any:
        schema = await self.getSchema(schemaId)
        schemaKey = schema.getKey()
        if schemaKey not in dictionary:
            raise ValueError(
                'No value for schema with ID={} and key={}'.format(
                    schemaId.schemaId, schemaId.schemaKey))
        return dictionary[schemaKey]

    async def _cacheValueForId(self, dictionary: Dict[SchemaKey, Any],
                               schemaId: ID, value: Any):
        schema = await self.getSchema(schemaId)
        schemaKey = schema.getKey()
        dictionary[schemaKey] = value



from typing import Sequence

from anoncreds.protocol.globals import PAIRING_GROUP
from anoncreds.protocol.revocation.accumulators.non_revocation_common import \
    createTauListExpectedValues, \
    createTauListValues
from anoncreds.protocol.types import T, NonRevocProof, ID, ProofRequest
from anoncreds.protocol.utils import int_to_ZR
from anoncreds.protocol.wallet.wallet import Wallet
from config.config import cmod


class NonRevocationProofVerifier:
    def __init__(self, wallet: Wallet):
        self._wallet = wallet

    async def verifyNonRevocation(self, proofRequest: ProofRequest, schema_seq_no,
                                  cHash, nonRevocProof: NonRevocProof) \
            -> Sequence[T]:
        if await self._wallet.shouldUpdateAccumulator(
                schemaId=ID(seqId=schema_seq_no),
                ts=proofRequest.ts,
                seqNo=proofRequest.seqNo):
            await self._wallet.updateAccumulator(schemaId=ID(schemaId=schema_seq_no),
                                                 ts=proofRequest.ts,
                                                 seqNo=proofRequest.seqNo)

        pkR = await self._wallet.getPublicKeyRevocation(ID(schemaId=schema_seq_no))
        accum = await self._wallet.getAccumulator(ID(schemaId=schema_seq_no))
        accumPk = await self._wallet.getPublicKeyAccumulator(ID(schemaId=schema_seq_no))

        CProof = nonRevocProof.CProof
        XList = nonRevocProof.XList

        group = cmod.PairingGroup(
            PAIRING_GROUP)  # super singular curve, 1024 bits
        THatExpected = createTauListExpectedValues(pkR, accum, accumPk, CProof)
        THatCalc = createTauListValues(pkR, accum, XList, CProof)
        chNum_z = int_to_ZR(cHash, group)

        return [(x ** chNum_z) * y for x, y in
                zip(THatExpected.asList(), THatCalc.asList())]



from functools import reduce

from anoncreds.protocol.globals import LARGE_NONCE
from anoncreds.protocol.primary.primary_proof_verifier import \
    PrimaryProofVerifier
from anoncreds.protocol.revocation.accumulators.non_revocation_proof_verifier import \
    NonRevocationProofVerifier
from anoncreds.protocol.types import FullProof, ProofRequest
from anoncreds.protocol.utils import get_hash_as_int, isCryptoInteger
from anoncreds.protocol.wallet.wallet import Wallet
from config.config import cmod


class Verifier:
    def __init__(self, wallet: Wallet):
        self.wallet = wallet
        self._primaryVerifier = PrimaryProofVerifier(wallet)
        self._nonRevocVerifier = NonRevocationProofVerifier(wallet)

    @property
    def verifierId(self):
        return self.wallet.walletId

    def generateNonce(self):
        return cmod.integer(cmod.randomBits(LARGE_NONCE))

    async def verify(self, proofRequest: ProofRequest, proof: FullProof):
        """
        Verifies a proof from the prover.

        :param proofRequest: description of a proof to be presented (revealed
        attributes, predicates, timestamps for non-revocation)
        :param proof: a proof
        :return: True if verified successfully and false otherwise.
        """

        if proofRequest.verifiableAttributes.keys() != proof.requestedProof.revealed_attrs.keys():
            raise ValueError('Received attributes ={} do not correspond to requested={}'.format(
                proof.requestedProof.revealed_attrs.keys(), proofRequest.verifiableAttributes.keys()))

        if proofRequest.predicates.keys() != proof.requestedProof.predicates.keys():
            raise ValueError('Received predicates ={} do not correspond to requested={}'.format(
                proof.requestedProof.predicates.keys(), proofRequest.predicates.keys()))

        TauList = []
        for (uuid, proofItem) in proof.proofs.items():
            if proofItem.proof.nonRevocProof:
                TauList += await self._nonRevocVerifier.verifyNonRevocation(
                    proofRequest, proofItem.schema_seq_no, proof.aggregatedProof.cHash,
                    proofItem.proof.nonRevocProof)
            if proofItem.proof.primaryProof:
                TauList += await self._primaryVerifier.verify(proofItem.schema_seq_no,
                                                              proof.aggregatedProof.cHash,
                                                              proofItem.proof.primaryProof)

        CHver = self._get_hash(proof.aggregatedProof.CList, self._prepare_collection(TauList),
                               cmod.integer(proofRequest.nonce))

        return CHver == proof.aggregatedProof.cHash

    def _prepare_collection(self, values):
        return [cmod.toInt(el) if isCryptoInteger(el) else el for el in values]

    def _get_hash(self, CList, TauList, nonce):
        return get_hash_as_int(nonce,
                               *reduce(lambda x, y: x + y, [TauList, CList]))



import pytest

from anoncreds.protocol.issuer import Issuer
from anoncreds.protocol.prover import Prover
from anoncreds.protocol.repo.attributes_repo import AttributeRepoInMemory
from anoncreds.protocol.repo.public_repo import PublicRepoInMemory
from anoncreds.protocol.types import AttribDef, AttribType, ID, ProofRequest
from anoncreds.protocol.verifier import Verifier
from anoncreds.protocol.wallet.issuer_wallet import IssuerWalletInMemory
from anoncreds.protocol.wallet.prover_wallet import ProverWalletInMemory
from anoncreds.protocol.wallet.wallet import WalletInMemory
from config.config import cmod

#
primes = {
    "prime1":
        (cmod.integer(
            157329491389375793912190594961134932804032426403110797476730107804356484516061051345332763141806005838436304922612495876180233509449197495032194146432047460167589034147716097417880503952139805241591622353828629383332869425029086898452227895418829799945650973848983901459733426212735979668835984691928193677469),
         cmod.integer(
             151323892648373196579515752826519683836764873607632072057591837216698622729557534035138587276594156320800768525825023728398410073692081011811496168877166664537052088207068061172594879398773872352920912390983199416927388688319207946493810449203702100559271439586753256728900713990097168484829574000438573295723)), "prime2":
        (cmod.integer(
            150619677884468353208058156632953891431975271416620955614548039937246769610622017033385394658879484186852231469238992217246264205570458379437126692055331206248530723117202131739966737760399755490935589223401123762051823602343810554978803032803606907761937587101969193241921351011430750970746500680609001799529),
         cmod.integer(
             171590857568436644992359347719703764048501078398666061921719064395827496970696879481740311141148273607392657321103691543916274965279072000206208571551864201305434022165176563363954921183576230072812635744629337290242954699427160362586102068962285076213200828451838142959637006048439307273563604553818326766703))
}

GVT = AttribDef('gvt',
                [AttribType('name', encode=True),
                 AttribType('age', encode=False),
                 AttribType('height', encode=False),
                 AttribType('sex', encode=True)])
XYZCorp = AttribDef('xyz',
                    [AttribType('status', encode=True),
                     AttribType('period', encode=False)])
# NASEMP = GVT + XYZCorp
#
iA1 = 100
iA2 = 101
#
proverId1 = 222
proverId2 = 333
#
# verifierId1 = 555
#
L = 5


#
#
# ############ module scope
#
#

@pytest.fixture(scope="module")
def primes1():
    P_PRIME1, Q_PRIME1 = primes.get("prime1")
    return dict(p_prime=P_PRIME1, q_prime=Q_PRIME1)


#
#
@pytest.fixture(scope="module")
def primes2():
    P_PRIME2, Q_PRIME2 = primes.get("prime2")
    return dict(p_prime=P_PRIME2, q_prime=Q_PRIME2)


@pytest.fixture(scope="function")
def attrRepo():
    return AttributeRepoInMemory()


@pytest.fixture(scope="function")
def publicRepo():
    return PublicRepoInMemory()


@pytest.fixture(scope="function")
def issuerWallet1(publicRepo):
    return IssuerWalletInMemory('issuer1', publicRepo)


@pytest.fixture(scope="function")
def issuerWallet2(publicRepo):
    return IssuerWalletInMemory('issuer2', publicRepo)


@pytest.fixture(scope="function")
def issuerGvt(issuerWallet1, attrRepo):
    return Issuer(issuerWallet1, attrRepo)


@pytest.fixture(scope="function")
def issuerXyz(issuerWallet2, attrRepo):
    return Issuer(issuerWallet2, attrRepo)


@pytest.fixture(scope="function")
def proverWallet1(publicRepo):
    return ProverWalletInMemory(proverId1, publicRepo)


@pytest.fixture(scope="function")
def proverWallet2(publicRepo):
    return ProverWalletInMemory(proverId2, publicRepo)


@pytest.fixture(scope="function")
def prover1(proverWallet1):
    return Prover(proverWallet1)


@pytest.fixture(scope="function")
def prover2(proverWallet2):
    return Prover(proverWallet2)


@pytest.fixture(scope="function")
def verifier(publicRepo):
    return Verifier(WalletInMemory('verifier1', publicRepo))


@pytest.fixture(scope="function")
def schemaGvt(issuerGvt, event_loop):
    return event_loop.run_until_complete(
        issuerGvt.genSchema("GVT", "1.0", GVT.attribNames()))


@pytest.fixture(scope="function")
def schemaXyz(issuerXyz, event_loop):
    return event_loop.run_until_complete(
        issuerXyz.genSchema("XYZCorp", "1.0", XYZCorp.attribNames()))


@pytest.fixture(scope="function")
def schemaGvtId(schemaGvt):
    return ID(schemaGvt.getKey())


@pytest.fixture(scope="function")
def schemaXyzId(schemaXyz):
    return ID(schemaXyz.getKey())


@pytest.fixture(scope="function")
def keysGvt(primes1, issuerGvt, schemaGvtId, event_loop):
    return event_loop.run_until_complete(issuerGvt.genKeys(schemaGvtId,
                                                           **primes1))


@pytest.fixture(scope="function")
def keysXyz(primes2, issuerXyz, schemaXyzId, event_loop):
    return event_loop.run_until_complete(issuerXyz.genKeys(schemaXyzId,
                                                           **primes2))


@pytest.fixture(scope="function")
def issueAccumulatorGvt(schemaGvtId, issuerGvt, keysGvt, event_loop):
    event_loop.run_until_complete(
        issuerGvt.issueAccumulator(schemaId=schemaGvtId, iA=iA1, L=L))


@pytest.fixture(scope="function")
def issueAccumulatorXyz(schemaXyzId, issuerXyz, keysXyz, event_loop):
    event_loop.run_until_complete(
        issuerXyz.issueAccumulator(schemaId=schemaXyzId, iA=iA2, L=L))


@pytest.fixture(scope="function")
def attrsProver1Gvt(attrRepo, schemaGvt):
    attrs = GVT.attribs(name='Alex', age=28, height=175, sex='male')
    attrRepo.addAttributes(schemaGvt.getKey(), proverId1, attrs)
    return attrs


@pytest.fixture(scope="function")
def attrsProver2Gvt(attrRepo, schemaGvt):
    attrs = GVT.attribs(name='Jason', age=42, height=180, sex='male')
    attrRepo.addAttributes(schemaGvt.getKey(), proverId2, attrs)
    return attrs


@pytest.fixture(scope="function")
def attrsProver1Xyz(attrRepo, schemaXyz):
    attrs = XYZCorp.attribs(status='partial', period=8)
    attrRepo.addAttributes(schemaXyz.getKey(), proverId1, attrs)
    return attrs


@pytest.fixture(scope="function")
def attrsProver2Xyz(attrRepo, schemaXyz):
    attrs = XYZCorp.attribs(status='full-time', period=22)
    attrRepo.addAttributes(schemaXyz.getKey(), proverId2, attrs)
    return attrs


@pytest.fixture(scope="function")
def claimsRequestProver1Gvt(prover1, schemaGvtId, keysGvt,
                            issueAccumulatorGvt, event_loop):
    return event_loop.run_until_complete(
        prover1.createClaimRequest(schemaGvtId))


@pytest.fixture(scope="function")
def claimsProver1Gvt(prover1, issuerGvt, claimsRequestProver1Gvt, schemaGvtId,
                     attrsProver1Gvt, event_loop):
    signature, claim = event_loop.run_until_complete(
        issuerGvt.issueClaim(schemaGvtId, claimsRequestProver1Gvt))
    event_loop.run_until_complete(
        prover1.processClaim(schemaGvtId, claim, signature))
    return event_loop.run_until_complete(
        prover1.wallet.getClaimSignature(schemaGvtId))


@pytest.fixture(scope="function")
def claimsProver2Gvt(prover2, issuerGvt, schemaGvtId, attrsProver2Gvt,
                     keysGvt, issueAccumulatorGvt, event_loop):
    claimsReq = event_loop.run_until_complete(
        prover2.createClaimRequest(schemaGvtId))
    signature, claim = event_loop.run_until_complete(
        issuerGvt.issueClaim(schemaGvtId, claimsReq))
    event_loop.run_until_complete(
        prover2.processClaim(schemaGvtId, claim, signature))
    return event_loop.run_until_complete(
        prover2.wallet.getClaimSignature(schemaGvtId))


@pytest.fixture(scope="function")
def claimsProver1Xyz(prover1, issuerXyz, schemaXyzId, attrsProver1Xyz,
                     keysXyz, issueAccumulatorXyz, event_loop):
    claimsReq = event_loop.run_until_complete(
        prover1.createClaimRequest(schemaXyzId))
    signature, claim = event_loop.run_until_complete(
        issuerXyz.issueClaim(schemaXyzId, claimsReq))
    event_loop.run_until_complete(
        prover1.processClaim(schemaXyzId, claim, signature))
    return event_loop.run_until_complete(
        prover1.wallet.getClaimSignature(schemaXyzId))


@pytest.fixture(scope="function")
def claimsProver2Xyz(prover2, issuerXyz, schemaXyzId, attrsProver2Xyz,
                     keysXyz, issueAccumulatorXyz, event_loop):
    claimsReq = event_loop.run_until_complete(
        prover2.createClaimRequest(schemaXyzId))
    signature, claim = event_loop.run_until_complete(
        issuerXyz.issueClaim(schemaXyzId, claimsReq))
    event_loop.run_until_complete(
        prover2.processClaim(schemaXyzId, claim, signature))
    return event_loop.run_until_complete(
        prover2.wallet.getClaimSignature(schemaXyzId))


@pytest.fixture(scope="function")
def claimsProver1(claimsProver1Gvt, claimsProver1Xyz):
    return claimsProver1Gvt, claimsProver1Xyz


@pytest.fixture(scope="function")
def claimsProver2(claimsProver2Gvt, claimsProver2Xyz):
    return claimsProver2Gvt, claimsProver2Xyz


@pytest.fixture(scope="function")
def allClaims(claimsProver1, claimsProver2):
    return claimsProver1, claimsProver2


@pytest.fixture(scope="function")
def nonce(verifier):
    return verifier.generateNonce()


@pytest.fixture(scope="function")
def genNonce(verifier):
    return verifier.generateNonce()


async def presentProofAndVerify(verifier: Verifier, proofRequest: ProofRequest, prover):
    proof = await prover.presentProof(proofRequest)
    return await verifier.verify(proofRequest, proof)



from anoncreds.protocol.globals import PAIRING_GROUP
from anoncreds.protocol.types import NonRevocationClaim, RevocationPublicKey, \
    RevocationSecretKey, \
    Accumulator, AccumulatorPublicKey, AccumulatorSecretKey, Witness, \
    ID, TimestampType, Tails
from anoncreds.protocol.utils import currentTimestampMillisec, groupIdentityG2
from anoncreds.protocol.wallet.issuer_wallet import IssuerWallet
from config.config import cmod


class NonRevocationClaimIssuer:
    def __init__(self, wallet: IssuerWallet):
        self._wallet = wallet

    async def genRevocationKeys(self) -> (
            RevocationPublicKey, RevocationSecretKey):
        group = cmod.PairingGroup(
            PAIRING_GROUP)  # super singular curve, 1024 bits

        g = group.random(cmod.G1)
        gprime = group.random(cmod.G2)

        h = group.random(cmod.G1)  # random element of the group G
        h0 = group.random(cmod.G1)
        h1 = group.random(cmod.G1)
        h2 = group.random(cmod.G1)
        htilde = group.random(cmod.G1)

        u = group.random(cmod.G2)
        hhat = group.random(cmod.G2)

        qr = group.order()  # order q_R of the group

        x = group.random(cmod.ZR)  # random(qr)
        sk = group.random(cmod.ZR)  # random(qr)

        pk = g ** sk
        y = hhat ** x

        return (RevocationPublicKey(qr, g, gprime, h, h0, h1, h2, htilde, hhat, u, pk, y),
                RevocationSecretKey(x, sk))

    async def issueAccumulator(self, schemaId, iA, L) \
            -> (Accumulator, Tails, AccumulatorPublicKey,
                AccumulatorSecretKey):
        pkR = await self._wallet.getPublicKeyRevocation(schemaId)
        group = cmod.PairingGroup(PAIRING_GROUP)
        gamma = group.random(cmod.ZR)

        tails = Tails()
        gCount = 2 * L
        for i in range(gCount):
            if i != L + 1:
                gVal = pkR.g ** (gamma ** i)
                gPrimeVal = pkR.gprime ** (gamma ** i)
                tails.addValue(i, gVal, gPrimeVal)
        z = cmod.pair(pkR.g, pkR.gprime) ** (gamma ** (L + 1))

        acc = 1
        V = set()

        accPK = AccumulatorPublicKey(z)
        accSK = AccumulatorSecretKey(gamma)
        accum = Accumulator(iA, acc, V, L)
        return accum, tails, accPK, accSK

    async def issueNonRevocationClaim(self, schemaId: ID, Ur, iA, i) -> (
            NonRevocationClaim, Accumulator, TimestampType):
        accum = await self._wallet.getAccumulator(schemaId)
        pkR = await self._wallet.getPublicKeyRevocation(schemaId)
        skR = await self._wallet.getSecretKeyRevocation(schemaId)
        tails = await self._wallet.getTails(schemaId)
        skAccum = await self._wallet.getSecretKeyAccumulator(schemaId)
        m2 = await self._wallet.getContextAttr(schemaId)

        if accum.isFull():
            raise ValueError("Accumulator is full. New one must be issued.")

        # TODO: currently all revo creds are issued sequentially
        group = cmod.PairingGroup(
            PAIRING_GROUP)  # super singular curve, 1024 bits

        i = i if i else accum.currentI
        accum.currentI += 1
        vrPrimeprime = group.random(cmod.ZR)
        c = group.random(cmod.ZR)

        m2 = group.init(cmod.ZR, int(m2))
        sigma = (pkR.h0 * (pkR.h1 ** m2) * Ur * tails.g[i] * (
            pkR.h2 ** vrPrimeprime)) ** (1 / (skR.x + c))
        omega = groupIdentityG2()
        for j in accum.V:
            omega *= tails.gprime[accum.L + 1 - j + i]

        sigmai = pkR.gprime ** (1 / (skR.sk + (skAccum.gamma ** i)))
        ui = pkR.u ** (skAccum.gamma ** i)

        accum.acc *= tails.gprime[accum.L + 1 - i]
        accum.V.add(i)

        witness = Witness(sigmai, ui, tails.g[i], omega, accum.V.copy())

        ts = currentTimestampMillisec()
        return (
            NonRevocationClaim(accum.iA, sigma, c, vrPrimeprime, witness,
                               i,
                               m2), accum, ts)

    async def revoke(self, schemaId: ID, i) -> (Accumulator, TimestampType):
        accum = await self._wallet.getAccumulator(schemaId)
        tails = await self._wallet.getTails(schemaId)

        accum.V.discard(i)
        accum.acc /= tails.gprime[accum.L + 1 - i]

        ts = currentTimestampMillisec()

        return accum, ts



import socket
import json
import asyncio
import logging

from anoncreds.protocol.issuer import Issuer
from anoncreds.protocol.repo.attributes_repo import AttributeRepoInMemory
from anoncreds.protocol.repo.public_repo import PublicRepoInMemory
from anoncreds.protocol.types import PredicateGE, \
    ID, AttributeInfo, ProofRequest, FullProof, PublicKey
from anoncreds.protocol.verifier import Verifier
from anoncreds.protocol.wallet.issuer_wallet import IssuerWalletInMemory
from anoncreds.protocol.wallet.wallet import WalletInMemory
from anoncreds.test.conftest import GVT

logging.basicConfig(format=u'%(filename)s[LINE:%(lineno)d]# %(levelname)-8s [%(asctime)s]  %(message)s',
                    level=logging.DEBUG)

loop = asyncio.get_event_loop()
global_dict = {
    'verifier': '',
    'proof_request': '',
    'public_key': ''
}
ip = '127.0.0.1'
port = 1234
chunk_size = 102400


def main():
    sock = socket.socket()
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((ip, port))
    sock.listen(1)
    logging.debug('Listening')
    conn, _ = sock.accept()
    logging.debug('Connected')

    while True:
        data = json.loads(conn.recv(chunk_size).decode("utf-8"))
        logging.debug('received data: {}'.format(data))
        if ('type' in data) & (data['type'] == 'receive_claim_def'):
            logging.debug('receive_claim_def -> start')
            global global_dict
            global_dict['public_key'] = PublicKey.from_str_dict(
                data['data']['data']['primary'])
            logging.debug('receive_claim_def -> done')
        if ('type' in data) & (data['type'] == 'get_proof_request'):
            logging.debug('get_proof_request -> start')
            create_request = asyncio.ensure_future(create_proof_request(conn))
            loop.run_until_complete(create_request)
            logging.debug('get_proof_request -> done')
        if ('type' in data) & (data['type'] == 'check_proof'):
            logging.debug('check_proof -> start')
            check_proof = asyncio.ensure_future(verify(data['data'], conn))
            loop.run_until_complete(check_proof)
            logging.debug('check_proof -> done')
        if (('type' in data) & (data['type'] == 'close')) | (not data):
            break

    sock.close()


async def create_proof_request(conn):
    # 1. Init entities
    public_repo = PublicRepoInMemory()
    attr_repo = AttributeRepoInMemory()
    issuer = Issuer(IssuerWalletInMemory('issuer1', public_repo), attr_repo)

    # 2. Create a Schema
    schema = await issuer.genSchema('GVT', '1.0', GVT.attribNames())
    schema_id = ID(schema.getKey())

    # 3. Create keys for the Schema
    global global_dict
    await issuer.wallet.submitPublicKeys(schema_id, global_dict['public_key'])

    # 4. set attributes for user1
    prover_id = 'BzfFCYk'
    attributes = GVT.attribs(name='Alex', age=28, height=175, sex='male')
    attr_repo.addAttributes(schema.getKey(), prover_id, attributes)

    verifier = Verifier(WalletInMemory('verifier1', public_repo))

    proof_request = ProofRequest(
        name='Test_proof', version='1.0',
        nonce=verifier.generateNonce(),
        verifiableAttributes={
            'attr_uuid': AttributeInfo('name', schema.seqId)},
        predicates={'predicate_uuid': PredicateGE('age', 18)})

    global_dict['verifier'] = verifier
    global_dict['proof_request'] = proof_request

    conn.send(json.dumps(proof_request.to_str_dict()).encode())


async def verify(proof, conn):
    proof = FullProof.from_str_dict(proof, [global_dict['public_key'].N])

    assert proof.requestedProof.revealed_attrs['attr_uuid'][1] == 'Alex'
    valid = await global_dict['verifier'].verify(global_dict['proof_request'], proof)

    conn.send(json.dumps(valid).encode())

if __name__ == '__main__':
    main()



from abc import abstractmethod
from typing import Any, Dict, Sequence

from anoncreds.protocol.repo.public_repo import PublicRepo
from anoncreds.protocol.types import Schema, SchemaKey, \
    PublicKey, ID, \
    RevocationPublicKey, AccumulatorPublicKey, Accumulator, Tails


class Wallet:
    def __init__(self, schemaId, repo: PublicRepo):
        self.walletId = schemaId
        self._repo = repo

    @property
    def name(self):
        return self.walletId

    # GET

    @abstractmethod
    async def getSchema(self, schemaId: ID) -> Schema:
        raise NotImplementedError

    @abstractmethod
    async def getAllSchemas(self) -> Sequence[Schema]:
        raise NotImplementedError

    @abstractmethod
    async def getPublicKey(self, schemaId: ID) -> PublicKey:
        raise NotImplementedError

    @abstractmethod
    async def getPublicKeyRevocation(self,
                                     schemaId: ID) -> RevocationPublicKey:
        raise NotImplementedError

    @abstractmethod
    async def getPublicKeyAccumulator(self,
                                      schemaId: ID) -> AccumulatorPublicKey:
        raise NotImplementedError

    @abstractmethod
    async def getAccumulator(self, schemaId: ID) -> Accumulator:
        raise NotImplementedError

    @abstractmethod
    async def updateAccumulator(self, schemaId: ID, ts=None, seqNo=None):
        raise NotImplementedError

    @abstractmethod
    async def shouldUpdateAccumulator(self, schemaId: ID, ts=None,
                                      seqNo=None):
        raise NotImplementedError

    @abstractmethod
    async def getTails(self, schemaId: ID) -> Tails:
        raise NotImplementedError


class WalletInMemory(Wallet):
    def __init__(self, schemaId, repo: PublicRepo):
        Wallet.__init__(self, schemaId, repo)

        # schema dicts
        self._schemasByKey = {}
        self._schemasById = {}

        # other dicts with key=schemaKey
        self._pks = {}
        self._pkRs = {}
        self._accums = {}
        self._accumPks = {}
        self._tails = {}

    # GET

    async def getSchema(self, schemaId: ID) -> Schema:
        if schemaId.schemaKey and schemaId.schemaKey in self._schemasByKey:
            return self._schemasByKey[schemaId.schemaKey]
        if schemaId.schemaId and schemaId.schemaId in self._schemasById:
            return self._schemasById[schemaId.schemaId]

        schema = await self._repo.getSchema(schemaId)

        self._cacheSchema(schema)

        return schema

    async def getAllSchemas(self) -> Sequence[Schema]:
        return self._schemasByKey.values()

    async def getPublicKey(self, schemaId: ID) -> PublicKey:
        return await self._getValueForId(self._pks, schemaId,
                                         self._repo.getPublicKey)

    async def getPublicKeyRevocation(self,
                                     schemaId: ID) -> RevocationPublicKey:
        return await self._getValueForId(self._pkRs, schemaId,
                                         self._repo.getPublicKeyRevocation)

    async def getPublicKeyAccumulator(self,
                                      schemaId: ID) -> AccumulatorPublicKey:
        return await self._getValueForId(self._accumPks, schemaId,
                                         self._repo.getPublicKeyAccumulator)

    async def getAccumulator(self, schemaId: ID) -> Accumulator:
        return await self._getValueForId(self._accums, schemaId,
                                         self._repo.getAccumulator)

    async def getTails(self, schemaId: ID) -> Tails:
        return await self._getValueForId(self._tails, schemaId,
                                         self._repo.getTails)

    async def updateAccumulator(self, schemaId: ID, ts=None, seqNo=None):
        acc = await self._repo.getAccumulator(schemaId)
        await self._cacheValueForId(self._accums, schemaId, acc)

    async def shouldUpdateAccumulator(self, schemaId: ID, ts=None,
                                      seqNo=None):
        # TODO
        return True

    # HELPER

    async def _getValueForId(self, dictionary: Dict[SchemaKey, Any],
                             schemaId: ID,
                             getFromRepo=None) -> Any:
        schema = await self.getSchema(schemaId)
        schemaKey = schema.getKey()

        if schemaKey in dictionary:
            return dictionary[schemaKey]

        elif getFromRepo:
            schemaId = schemaId._replace(schemaKey=schemaKey,
                                         schemaId=schema.seqId)
            value = await getFromRepo(schemaId)
            dictionary[schemaKey] = value
            return value

        else:
            raise ValueError(
                'No value for schema with ID={} and key={}'.format(
                    schemaId.schemaId, schemaId.schemaKey))

    async def _cacheValueForId(self, dictionary: Dict[SchemaKey, Any],
                               schemaId: ID, value: Any):
        schema = await self.getSchema(schemaId)
        schemaKey = schema.getKey()
        dictionary[schemaKey] = value

    def _cacheSchema(self, schema: Schema):
        self._schemasByKey[schema.getKey()] = schema
        if schema.seqId:
            self._schemasById[schema.seqId] = schema



LARGE_PUBLIC_RHO = 256
LARGE_PUBLIC_B = 1376
LARGE_PRIME = 1024
LARGE_VPRIME_PRIME = 2724
LARGE_VPRIME = 2128
LARGE_MVECT = 592
LARGE_UTILDE = LARGE_MVECT
LARGE_RTILDE = 672
LARGE_E_START = 596
LARGE_E_END_RANGE = 119
LARGE_NONCE = 80
LARGE_ETILDE = 456
LARGE_VTILDE = 3060
LARGE_MASTER_SECRET = 256
LARGE_ALPHATILDE = 2787
LARGE_M2_TILDE = 1024
ITERATIONS = 4

PAIRING_GROUP = 'SS1024'  # super singular curve, 1024 bits

MASTER_SEC_RAND = "master_secret_rand"
REVEALED_ATTRS = "revealedAttrs"
ZERO_INDEX = "0"

ATTRS = "attrs"
DELTA = "DELTA"
TVAL = "Tval"
KEYS = "keys"
NONCE = "nonce"

APRIME = "Aprime"
EPRIME = "eprime"
VPRIME = "vprime"

EVECT = "evect"
MVECT = "mvect"
VVECT = "vvect"

ETILDE = "etilde"
MTILDE = "mtilde"
VTILDE = "vtilde"

CRED_A = "A"
CRED_E = "e"
CRED_V = "v"

C_VALUE = "c"

PK_R = "R"
PK_N = "N"
PK_S = "S"
PK_Z = "Z"

ISSUER = "issuer"
PROOF = "proof"

NAME = "name"
VERSION = "version"
TYPE = "type"
ATTR_NAMES = "attr_names"

TYPE_CL = "CL"



from abc import abstractmethod

from anoncreds.protocol.repo.public_repo import PublicRepo
from anoncreds.protocol.types import Schema, PublicKey, SecretKey, ID, \
    RevocationPublicKey, AccumulatorPublicKey, Accumulator, Tails, \
    RevocationSecretKey, AccumulatorSecretKey, \
    TimestampType
from anoncreds.protocol.wallet.wallet import Wallet, WalletInMemory


class IssuerWallet(Wallet):
    def __init__(self, schemaId, repo: PublicRepo):
        Wallet.__init__(self, schemaId, repo)

    # SUBMIT

    @abstractmethod
    async def submitSchema(self,
                           schema: Schema) -> Schema:
        raise NotImplementedError

    @abstractmethod
    async def submitPublicKeys(self, schemaId: ID, pk: PublicKey,
                               pkR: RevocationPublicKey = None) -> (
            PublicKey, RevocationPublicKey):
        raise NotImplementedError

    @abstractmethod
    async def submitSecretKeys(self, schemaId: ID, sk: SecretKey,
                               skR: RevocationSecretKey = None):
        raise NotImplementedError

    @abstractmethod
    async def submitAccumPublic(self, schemaId: ID,
                                accumPK: AccumulatorPublicKey,
                                accum: Accumulator, tails: Tails):
        raise NotImplementedError

    @abstractmethod
    async def submitAccumSecret(self, schemaId: ID,
                                accumSK: AccumulatorSecretKey):
        raise NotImplementedError

    @abstractmethod
    async def submitAccumUpdate(self, schemaId: ID, accum: Accumulator,
                                timestampMs: TimestampType):
        raise NotImplementedError

    @abstractmethod
    async def submitContextAttr(self, schemaId: ID, m2):
        raise NotImplementedError

    # GET

    @abstractmethod
    async def getSecretKey(self, schemaId: ID) -> SecretKey:
        raise NotImplementedError

    @abstractmethod
    async def getSecretKeyRevocation(self,
                                     schemaId: ID) -> RevocationSecretKey:
        raise NotImplementedError

    @abstractmethod
    async def getSecretKeyAccumulator(self,
                                      schemaId: ID) -> AccumulatorSecretKey:
        raise NotImplementedError

    @abstractmethod
    async def getContextAttr(self, schemaId: ID):
        raise NotImplementedError


class IssuerWalletInMemory(IssuerWallet, WalletInMemory):
    def __init__(self, schemaId, repo: PublicRepo):
        WalletInMemory.__init__(self, schemaId, repo)

        # other dicts with key=schemaKey
        self._sks = {}
        self._skRs = {}
        self._accumSks = {}
        self._m2s = {}
        self._attributes = {}

    # SUBMIT

    async def submitSchema(self,
                           schema: Schema) -> Schema:
        schema = await self._repo.submitSchema(schema)
        if schema:
            self._cacheSchema(schema)
        return schema

    async def submitPublicKeys(self, schemaId: ID, pk: PublicKey,
                               pkR: RevocationPublicKey = None) -> (
            PublicKey, RevocationPublicKey):
        pk, pkR = await self._repo.submitPublicKeys(schemaId, pk, pkR)
        await self._cacheValueForId(self._pks, schemaId, pk)
        if pkR:
            await  self._cacheValueForId(self._pkRs, schemaId, pkR)
        return pk, pkR

    async def submitSecretKeys(self, schemaId: ID, sk: SecretKey,
                               skR: RevocationSecretKey = None):
        await  self._cacheValueForId(self._sks, schemaId, sk)
        if skR:
            await  self._cacheValueForId(self._skRs, schemaId, skR)

    async def submitAccumPublic(self, schemaId: ID,
                                accumPK: AccumulatorPublicKey,
                                accum: Accumulator,
                                tails: Tails) -> AccumulatorPublicKey:
        accumPK = await self._repo.submitAccumulator(schemaId, accumPK, accum,
                                                     tails)
        await self._cacheValueForId(self._accums, schemaId, accum)
        await self._cacheValueForId(self._accumPks, schemaId, accumPK)
        await self._cacheValueForId(self._tails, schemaId, tails)
        return accumPK

    async def submitAccumSecret(self, schemaId: ID,
                                accumSK: AccumulatorSecretKey):
        await self._cacheValueForId(self._accumSks, schemaId, accumSK)

    async def submitAccumUpdate(self, schemaId: ID, accum: Accumulator,
                                timestampMs: TimestampType):
        await self._repo.submitAccumUpdate(schemaId, accum, timestampMs)
        await self._cacheValueForId(self._accums, schemaId, accum)

    async def submitContextAttr(self, schemaId: ID, m2):
        await self._cacheValueForId(self._m2s, schemaId, m2)

    # GET

    async def getSecretKey(self, schemaId: ID) -> SecretKey:
        return await self._getValueForId(self._sks, schemaId)

    async def getSecretKeyRevocation(self,
                                     schemaId: ID) -> RevocationSecretKey:
        return await self._getValueForId(self._skRs, schemaId)

    async def getSecretKeyAccumulator(self,
                                      schemaId: ID) -> AccumulatorSecretKey:
        return await self._getValueForId(self._accumSks, schemaId)

    async def getContextAttr(self, schemaId: ID):
        return await self._getValueForId(self._m2s, schemaId)



import os
import sys

from setuptools import setup, find_packages, __version__
from anoncreds.__metadata__ import __author__, __license__

v = sys.version_info
if sys.version_info < (3, 5):
    msg = "FAIL: Requires Python 3.5 or later, " \
          "but setup.py was run using {}.{}.{}"
    v = sys.version_info
    print(msg.format(v.major, v.minor, v.micro))
    print("NOTE: Installation failed. Run setup.py using python3")
    sys.exit(1)

# Change to ioflo's source directory prior to running any command
try:
    SETUP_DIRNAME = os.path.dirname(__file__)
except NameError:
    # We're probably being frozen, and __file__ triggered this NameError
    # Work around this
    SETUP_DIRNAME = os.path.dirname(sys.argv[0])

if SETUP_DIRNAME != '':
    os.chdir(SETUP_DIRNAME)

SETUP_DIRNAME = os.path.abspath(SETUP_DIRNAME)

METADATA = os.path.join(SETUP_DIRNAME, 'anoncreds', '__metadata__.py')
# Load the metadata using exec() so we don't trigger an import of ioflo.__init__
exec(compile(open(METADATA).read(), METADATA, 'exec'))

setup(
    name='indy-anoncreds-dev',
    version=__version__,
    description='Anonymous credentials',
    long_description='Anonymous credentials',
    url='https://github.com/hyperledger/indy-anoncreds',
    author=__author__,
    author_email='hyperledger-indy@lists.hyperledger.org',
    license=__license__,
    keywords='Anonymous credentials',
    packages=find_packages(exclude=['docs', 'docs*']),
    package_data={
        '': ['*.txt', '*.md', '*.rst', '*.json', '*.conf', '*.html',
             '*.css', '*.ico', '*.png', 'LICENSE', 'LEGAL']},
    install_requires=['Charm-Crypto', 'lazy-object-proxy', 'base58'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-asyncio']
)



from functools import reduce
from typing import Dict, Sequence, Any

from anoncreds.protocol.globals import LARGE_MASTER_SECRET, LARGE_M2_TILDE
from anoncreds.protocol.primary.primary_proof_builder import \
    PrimaryClaimInitializer, PrimaryProofBuilder
from anoncreds.protocol.revocation.accumulators.non_revocation_proof_builder import \
    NonRevocationClaimInitializer, \
    NonRevocationProofBuilder
from anoncreds.protocol.types import PrimaryClaim, NonRevocationClaim, Proof, \
    InitProof, ProofRequest, ProofClaims, \
    FullProof, \
    Schema, ID, SchemaKey, ClaimRequest, Claims, RequestedProof, AggregatedProof, ProofInfo, ClaimAttributeValues
from anoncreds.protocol.utils import get_hash_as_int, isCryptoInteger
from anoncreds.protocol.wallet.prover_wallet import ProverWallet
from config.config import cmod


class Prover:
    def __init__(self, wallet: ProverWallet):
        self.wallet = wallet

        self._primaryClaimInitializer = PrimaryClaimInitializer(wallet)
        self._nonRevocClaimInitializer = NonRevocationClaimInitializer(wallet)

        self._primaryProofBuilder = PrimaryProofBuilder(wallet)
        self._nonRevocProofBuilder = NonRevocationProofBuilder(wallet)

    #
    # PUBLIC
    #

    @property
    def proverId(self):
        return self.wallet.walletId

    async def createClaimRequest(self, schemaId: ID, proverId=None,
                                 reqNonRevoc=True) -> ClaimRequest:
        """
        Creates a claim request to the issuer.

        :param schemaId: The schema ID (reference to claim
        definition schema)
        :param proverId: a prover ID request a claim for (if None then
        the current prover default ID is used)
        :param reqNonRevoc: whether to request non-revocation claim
        :return: Claim Request
        """
        await self._genMasterSecret(schemaId)
        U = await self._genU(schemaId)
        Ur = None if not reqNonRevoc else await self._genUr(schemaId)
        proverId = proverId if proverId else self.proverId
        return ClaimRequest(userId=proverId, U=U, Ur=Ur)

    async def createClaimRequests(self, schemaIds: Sequence[ID],
                                  proverId=None,
                                  reqNonRevoc=True) -> Dict[ID, ClaimRequest]:
        """
        Creates a claim request to the issuer.

        :param schemaIds: The schema IDs (references to claim
        definition schema)
        :param proverId: a prover ID request a claim for (if None then
        the current prover default ID is used)
        :param reqNonRevoc: whether to request non-revocation claim
        :return: a dictionary of Claim Requests for each Schema.
        """
        res = {}
        for schemaId in schemaIds:
            res[schemaId] = await self.createClaimRequest(schemaId,
                                                          proverId,
                                                          reqNonRevoc)
        return res

    async def processClaim(self, schemaId: ID, claimAttributes: Dict[str, ClaimAttributeValues], signature: Claims):
        """
        Processes and saves a received Claim for the given Schema.

        :param schemaId: The schema ID (reference to claim
        definition schema)
        :param claims: claims to be processed and saved
        """
        await self.wallet.submitContextAttr(schemaId, signature.primaryClaim.m2)
        await self.wallet.submitClaimAttributes(schemaId, claimAttributes)

        await self._initPrimaryClaim(schemaId, signature.primaryClaim)
        if signature.nonRevocClaim:
            await self._initNonRevocationClaim(schemaId, signature.nonRevocClaim)

    async def processClaims(self, allClaims: Dict[ID, Claims]):
        """
        Processes and saves received Claims.

        :param claims: claims to be processed and saved for each claim
        definition.
        """
        res = []
        for schemaId, (claim_signature, claim_attributes) in allClaims.items():
            res.append(await self.processClaim(schemaId, claim_attributes, claim_signature))
        return res

    async def presentProof(self, proofRequest: ProofRequest) -> FullProof:
        """
        Presents a proof to the verifier.

        :param proofRequest: description of a proof to be presented (revealed
        attributes, predicates, timestamps for non-revocation)
        :return: a proof (both primary and non-revocation) and revealed attributes (initial non-encoded values)
        """
        claims, requestedProof = await self._findClaims(proofRequest)
        proof = await self._prepareProof(claims, proofRequest.nonce, requestedProof)
        return proof

    #
    # REQUEST CLAIMS
    #

    async def _genMasterSecret(self, schemaId: ID):
        ms = cmod.integer(cmod.randomBits(LARGE_MASTER_SECRET))
        await self.wallet.submitMasterSecret(schemaId=schemaId, ms=ms)

    async def _genU(self, schemaId: ID):
        claimInitData = await self._primaryClaimInitializer.genClaimInitData(
            schemaId)
        await self.wallet.submitPrimaryClaimInitData(schemaId=schemaId,
                                                     claimInitData=claimInitData)
        return claimInitData.U

    async def _genUr(self, schemaId: ID):
        claimInitData = await self._nonRevocClaimInitializer.genClaimInitData(
            schemaId)
        await self.wallet.submitNonRevocClaimInitData(schemaId=schemaId,
                                                      claimInitData=claimInitData)
        return claimInitData.U

    async def _initPrimaryClaim(self, schemaId: ID, claim: PrimaryClaim):
        claim = await self._primaryClaimInitializer.preparePrimaryClaim(
            schemaId,
            claim)
        await self.wallet.submitPrimaryClaim(schemaId=schemaId, claim=claim)

    async def _initNonRevocationClaim(self, schemaId: ID,
                                      claim: NonRevocationClaim):
        claim = await self._nonRevocClaimInitializer.initNonRevocationClaim(
            schemaId,
            claim)
        await self.wallet.submitNonRevocClaim(schemaId=schemaId,
                                              claim=claim)

    #
    # PRESENT PROOF
    #

    async def _findClaims(self, proofRequest: ProofRequest) -> (
            Dict[SchemaKey, ProofClaims], Dict[str, Any]):
        revealedAttrs, predicates = proofRequest.verifiableAttributes, proofRequest.predicates

        foundRevealedAttrs = {}
        foundPredicates = {}
        proofClaims = {}
        schemas = {}
        allClaimsAttributes = await self.wallet.getAllClaimsAttributes()

        async def addProof():
            revealedAttrsForClaim = [
                a for a in revealedAttrs.values() if a.name in claim.keys()]
            revealedPredicatesForClaim = [
                p for p in predicates.values() if p.attrName in claim.keys()]

            claims = await self.wallet.getClaimSignature(ID(schemaId=schemaId))
            proofClaim = ProofClaims(claims=claims, revealedAttrs=revealedAttrsForClaim,
                                     predicates=revealedPredicatesForClaim)

            proofClaims[schemaId] = proofClaim

        for schemaKey, c in allClaimsAttributes.items():
            schemas[schemaKey] = (await self.wallet.getSchema(ID(schemaKey)))

        for uuid, revealedAttr in revealedAttrs.items():
            matches = [(schemas[key].seqId, c) for key, c in allClaimsAttributes.items() if revealedAttr.name in c and
                       (schemas[key].seqId == revealedAttr.schema_seq_no if revealedAttr.schema_seq_no else True) and
                       (schemas[key].issuerId == revealedAttr.issuer_did if revealedAttr.issuer_did else True)]

            if len(matches) == 0:
                raise ValueError(
                    "A claim isn't found for the following attributes: {}", revealedAttr.name)

            schemaId, claim = matches[0]
            foundRevealedAttrs[uuid] = [str(schemaId), str(claim[revealedAttr.name].raw),
                                        str(claim[revealedAttr.name].encoded)]

            if schemaId not in proofClaims:
                await addProof()

        for uuid, predicate in predicates.items():
            matches = [(schemas[key].seqId, c) for key, c in allClaimsAttributes.items() if predicate.attrName in c and
                       (schemas[key].seqId == predicate.schema_seq_no if predicate.schema_seq_no else True) and
                       (schemas[key].issuerId == predicate.issuer_did if predicate.issuer_did else True)]

            if len(matches) == 0:
                raise ValueError(
                    "A claim isn't found for the following predicate: {}", predicate)

            schemaId, claim = matches[0]
            foundPredicates[uuid] = str(schemaId)

            if schemaId not in proofClaims:
                await addProof()

        requestedProof = RequestedProof(
            revealed_attrs=foundRevealedAttrs, predicates=foundPredicates)

        return proofClaims, requestedProof

    async def _prepareProof(self, claims: Dict[SchemaKey, ProofClaims],
                            nonce, requestedProof) -> FullProof:
        m1Tilde = cmod.integer(cmod.randomBits(LARGE_M2_TILDE))
        initProofs = {}
        CList = []
        TauList = []

        # 1. init proofs
        for schemaId, val in claims.items():
            c1, c2, revealedAttrs, predicates = val.claims.primaryClaim, val.claims.nonRevocClaim, val.revealedAttrs, val.predicates

            claim = await self.wallet.getClaimAttributes(ID(schemaId=schemaId))

            nonRevocInitProof = None
            if c2:
                nonRevocInitProof = await self._nonRevocProofBuilder.initProof(
                    schemaId, c2)
                CList += nonRevocInitProof.asCList()
                TauList += nonRevocInitProof.asTauList()

            primaryInitProof = None
            if c1:
                m2Tilde = cmod.integer(int(
                    nonRevocInitProof.TauListParams.m2)) if nonRevocInitProof else None
                primaryInitProof = await self._primaryProofBuilder.initProof(
                    schemaId, c1, revealedAttrs, predicates,
                    m1Tilde, m2Tilde, claim)
                CList += primaryInitProof.asCList()
                TauList += primaryInitProof.asTauList()

            initProof = InitProof(nonRevocInitProof, primaryInitProof)
            initProofs[schemaId] = initProof

        # 2. hash
        cH = self._get_hash(self._prepare_collection(
            CList), self._prepare_collection(TauList), nonce)

        # 3. finalize proofs
        proofs = {}
        for schemaId, initProof in initProofs.items():
            nonRevocProof = None
            if initProof.nonRevocInitProof:
                nonRevocProof = await self._nonRevocProofBuilder.finalizeProof(
                    schemaId, cH, initProof.nonRevocInitProof)
            primaryProof = await self._primaryProofBuilder.finalizeProof(
                schemaId, cH, initProof.primaryInitProof)

            schema = await self.wallet.getSchema(ID(schemaId=schemaId))

            proof = Proof(primaryProof, nonRevocProof)
            proofInfo = ProofInfo(
                proof=proof, schema_seq_no=schemaId, issuer_did=schema.issuerId)

            proofs[str(schemaId)] = proofInfo

        aggregatedProof = AggregatedProof(cH, self._prepare_collection(CList))

        return FullProof(proofs, aggregatedProof, requestedProof)

    async def _getCList(self, initProofs: Dict[Schema, InitProof]):
        CList = []
        for initProof in initProofs.values():
            CList += await initProof.nonRevocInitProof.asCList()
            CList += await initProof.primaryInitProof.asCList()
            return CList

    async def _getTauList(self, initProofs: Dict[Schema, InitProof]):
        TauList = []
        for initProof in initProofs.values():
            TauList += await initProof.nonRevocInitProof.asTauList()
            TauList += await initProof.primaryInitProof.asTauList()
        return TauList

    def _prepare_collection(self, values):
        return [cmod.toInt(el) if isCryptoInteger(el) else el for el in values]

    def _get_hash(self, CList, TauList, nonce):
        return get_hash_as_int(nonce,
                               *reduce(lambda x, y: x + y, [TauList, CList]))



from anoncreds.protocol.globals import LARGE_E_START, ITERATIONS, DELTA
from anoncreds.protocol.primary.primary_proof_common import calcTeq, calcTge
from anoncreds.protocol.types import PrimaryEqualProof, \
    PrimaryPredicateGEProof, PrimaryProof, ID
from anoncreds.protocol.wallet.wallet import Wallet
from config.config import cmod


class PrimaryProofVerifier:
    def __init__(self, wallet: Wallet):
        self._wallet = wallet

    async def verify(self, schemaId, cHash, primaryProof: PrimaryProof):
        cH = cmod.integer(cHash)
        THat = await self._verifyEquality(schemaId, cH, primaryProof.eqProof)
        for geProof in primaryProof.geProofs:
            THat += await self._verifyGEPredicate(schemaId, cH, geProof)

        return THat

    async def _verifyEquality(self, schemaId, cH, proof: PrimaryEqualProof):
        THat = []
        pk = await self._wallet.getPublicKey(ID(schemaId=schemaId))
        attrNames = (await self._wallet.getSchema(ID(schemaId=schemaId))).attrNames
        unrevealedAttrNames = set(attrNames) - set(proof.revealedAttrs.keys())

        T1 = calcTeq(pk, proof.Aprime, proof.e, proof.v,
                     proof.m, proof.m1, proof.m2,
                     unrevealedAttrNames)

        Rar = 1 % pk.N
        for attrName in proof.revealedAttrs.keys():
            Rar *= pk.R[str(attrName)] ** proof.revealedAttrs[str(attrName)]
        Rar *= proof.Aprime ** (2 ** LARGE_E_START)
        T2 = (pk.Z / Rar) ** (-1 * cH) % pk.N
        T = T1 * T2 % pk.N

        THat.append(T)
        return THat

    async def _verifyGEPredicate(self, schemaId, cH,
                                 proof: PrimaryPredicateGEProof):
        pk = await self._wallet.getPublicKey(ID(schemaId=schemaId))
        v = proof.predicate.value

        TauList = calcTge(pk, proof.u, proof.r, proof.mj, proof.alpha, proof.T)

        for i in range(0, ITERATIONS):
            TT = proof.T[str(i)] ** (-1 * cH) % pk.N
            TauList[i] = TauList[i] * TT % pk.N
        TauList[ITERATIONS] = TauList[ITERATIONS] * (
            (proof.T[DELTA] * (pk.Z ** v)) ** (-1 * cH)) % pk.N
        TauList[ITERATIONS + 1] = (TauList[ITERATIONS + 1] * (
            proof.T[DELTA] ** (-1 * cH))) % pk.N

        return TauList



"""
This is an example of a crypto configuration file that loads in a library
dynamically.
"""

# noinspection PyUnresolvedReferences
from charm.core.math.integer import integer, random, randomBits, isPrime, \
    randomPrime, serialize, deserialize, toInt

# noinspection PyUnresolvedReferences
from charm.toolbox.conversion import Conversion

# noinspection PyUnresolvedReferences
from charm.toolbox.pairinggroup import PairingGroup, ZR, G1, G2, pair, pc_element



from anoncreds.protocol.types import Accumulator, NonRevocProofXList, \
    NonRevocProofCList, RevocationPublicKey, \
    NonRevocProofTauList, AccumulatorPublicKey
from anoncreds.protocol.utils import groupIdentityG1
from config.config import cmod


def createTauListValues(pk: RevocationPublicKey, accum: Accumulator,
                        params: NonRevocProofXList,
                        proofC: NonRevocProofCList) -> NonRevocProofTauList:
    T1 = (pk.h ** params.rho) * (pk.htilde ** params.o)
    T2 = (proofC.E ** params.c) * (pk.h ** (-params.m)) * (
        pk.htilde ** (-params.t))
    T3 = ((cmod.pair(proofC.A, pk.hhat) ** params.c) *
          (cmod.pair(pk.htilde, pk.hhat) ** params.r)) / \
         ((cmod.pair(pk.htilde, pk.y) ** params.rho) *
          (cmod.pair(pk.htilde, pk.hhat) ** params.m) *
          (cmod.pair(pk.h1, pk.hhat) ** params.m2) *
          (cmod.pair(pk.h2, pk.hhat) ** params.s))
    T4 = (cmod.pair(pk.htilde, accum.acc) ** params.r) * \
         (cmod.pair(1 / pk.g, pk.hhat) ** params.rPrime)
    T5 = (pk.g ** params.r) * (pk.htilde ** params.oPrime)
    T6 = (proofC.D ** params.rPrimePrime) * (pk.g ** -params.mPrime) * (
        pk.htilde ** -params.tPrime)
    T7 = (cmod.pair(pk.pk * proofC.G, pk.hhat) ** params.rPrimePrime) * \
         (cmod.pair(pk.htilde, pk.hhat) ** -params.mPrime) * \
         (cmod.pair(pk.htilde, proofC.S) ** params.r)
    T8 = (cmod.pair(pk.htilde, pk.u) ** params.r) * \
         (cmod.pair(1 / pk.g, pk.hhat) ** params.rPrimePrimePrime)
    return NonRevocProofTauList(T1, T2, T3, T4, T5, T6, T7, T8)


def createTauListExpectedValues(pk: RevocationPublicKey, accum: Accumulator,
                                accumPk: AccumulatorPublicKey,
                                proofC: NonRevocProofCList) -> NonRevocProofTauList:
    T1 = proofC.E
    T2 = groupIdentityG1()
    T3 = cmod.pair(pk.h0 * proofC.G, pk.hhat) / cmod.pair(proofC.A, pk.y)
    T4 = cmod.pair(proofC.G, accum.acc) / (
        cmod.pair(pk.g, proofC.W) * accumPk.z)
    T5 = proofC.D
    T6 = groupIdentityG1()
    T7 = cmod.pair(pk.pk * proofC.G, proofC.S) / cmod.pair(pk.g, pk.gprime)
    T8 = cmod.pair(proofC.G, pk.u) / cmod.pair(pk.g, proofC.U)
    return NonRevocProofTauList(T1, T2, T3, T4, T5, T6, T7, T8)



import pytest

from anoncreds.protocol.issuer import Issuer
from anoncreds.protocol.prover import Prover
from anoncreds.protocol.repo.attributes_repo import AttributeRepoInMemory
from anoncreds.protocol.repo.public_repo import PublicRepoInMemory
from anoncreds.protocol.types import ProofRequest, PredicateGE, \
    ID, AttributeInfo
from anoncreds.protocol.verifier import Verifier
from anoncreds.protocol.wallet.issuer_wallet import IssuerWalletInMemory
from anoncreds.protocol.wallet.prover_wallet import ProverWalletInMemory
from anoncreds.protocol.wallet.wallet import WalletInMemory
from anoncreds.test.conftest import GVT, XYZCorp


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testSingleIssuerSingleProver(primes1):
    # 1. Init entities
    publicRepo = PublicRepoInMemory()
    attrRepo = AttributeRepoInMemory()
    issuer = Issuer(IssuerWalletInMemory('issuer1', publicRepo), attrRepo)

    # 2. Create a Schema
    schema = await issuer.genSchema('GVT', '1.0', GVT.attribNames())
    schemaId = ID(schema.getKey())

    # 3. Create keys for the Schema
    await issuer.genKeys(schemaId, **primes1)

    # 4. Issue accumulator
    await issuer.issueAccumulator(schemaId=schemaId, iA='110', L=5)

    # 4. set attributes for user1
    userId = '111'
    attrs = GVT.attribs(name='Alex', age=28, height=175, sex='male')
    attrRepo.addAttributes(schema.getKey(), userId, attrs)

    # 5. request Claims
    prover = Prover(ProverWalletInMemory(userId, publicRepo))
    claimsReq = await prover.createClaimRequest(schemaId)
    (claim_signature, claim_attributes) = await issuer.issueClaim(schemaId, claimsReq)

    await prover.processClaim(schemaId, claim_attributes, claim_signature)

    # 6. proof Claims
    verifier = Verifier(WalletInMemory('verifier1', publicRepo))

    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo('name', schema.seqId)},
                                predicates={'predicate_uuid': PredicateGE('age', 18)})

    proof = await prover.presentProof(proofRequest)
    assert proof.requestedProof.revealed_attrs['attr_uuid'][1] == 'Alex'
    assert await verifier.verify(proofRequest, proof)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testMultiplIssuersSingleProver(primes1, primes2):
    # 1. Init entities
    publicRepo = PublicRepoInMemory()
    attrRepo = AttributeRepoInMemory()
    issuer1 = Issuer(IssuerWalletInMemory('issuer1', publicRepo), attrRepo)
    issuer2 = Issuer(IssuerWalletInMemory('issuer2', publicRepo), attrRepo)

    # 2. Create a Schema
    schema1 = await issuer1.genSchema('GVT', '1.0', GVT.attribNames())
    schemaId1 = ID(schema1.getKey())
    schema2 = await issuer2.genSchema('XYZCorp', '1.0',
                                      XYZCorp.attribNames())
    schemaId2 = ID(schema2.getKey())

    # 3. Create keys for the Schema
    await issuer1.genKeys(schemaId1, **primes1)
    await issuer2.genKeys(schemaId2, **primes2)

    # 4. Issue accumulator
    await issuer1.issueAccumulator(schemaId=schemaId1, iA='110', L=5)
    await issuer2.issueAccumulator(schemaId=schemaId2, iA=9999999, L=5)

    # 4. set attributes for user1
    userId = '111'
    attrs1 = GVT.attribs(name='Alex', age=28, height=175, sex='male')
    attrs2 = XYZCorp.attribs(status='FULL', period=8)
    attrRepo.addAttributes(schema1.getKey(), userId, attrs1)
    attrRepo.addAttributes(schema2.getKey(), userId, attrs2)

    # 5. request Claims
    prover = Prover(ProverWalletInMemory(userId, publicRepo))
    claimsReq1 = await prover.createClaimRequest(schemaId1)
    claimsReq2 = await prover.createClaimRequest(schemaId2)
    (claim1_signature, claim1_attributes) = await issuer1.issueClaim(schemaId1, claimsReq1)
    (claim2_signature, claim2_attributes) = await issuer2.issueClaim(schemaId2, claimsReq2)
    await prover.processClaim(schemaId1, claim1_attributes, claim1_signature)
    await prover.processClaim(schemaId2, claim2_attributes, claim2_signature)

    # 6. proof Claims
    verifier = Verifier(WalletInMemory('verifier1', publicRepo))

    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={'attr_uuid1': AttributeInfo('name', schema1.seqId),
                                                      'attr_uuid2': AttributeInfo('status', schema2.seqId)},
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('period', 5)})

    proof = await prover.presentProof(proofRequest)

    assert proof.requestedProof.revealed_attrs['attr_uuid1'][1] == 'Alex'
    assert proof.requestedProof.revealed_attrs['attr_uuid2'][1] == 'FULL'

    assert await verifier.verify(proofRequest, proof)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testSingleIssuerMultipleCredDefsSingleProver(primes1, primes2):
    # 1. Init entities
    publicRepo = PublicRepoInMemory()
    attrRepo = AttributeRepoInMemory()
    issuer = Issuer(IssuerWalletInMemory('issuer1', publicRepo), attrRepo)

    # 2. Create a Schema
    schema1 = await issuer.genSchema('GVT', '1.0', GVT.attribNames())
    schemaId1 = ID(schema1.getKey())
    schema2 = await issuer.genSchema('XYZCorp', '1.0',
                                     XYZCorp.attribNames())
    schemaId2 = ID(schema2.getKey())

    # 3. Create keys for the Schema
    await issuer.genKeys(schemaId1, **primes1)
    await issuer.genKeys(schemaId2, **primes2)

    # 4. Issue accumulator
    await issuer.issueAccumulator(schemaId=schemaId1, iA='110', L=5)
    await issuer.issueAccumulator(schemaId=schemaId2, iA=9999999, L=5)

    # 4. set attributes for user1
    userId = '111'
    attrs1 = GVT.attribs(name='Alex', age=28, height=175, sex='male')
    attrs2 = XYZCorp.attribs(status='FULL', period=8)
    attrRepo.addAttributes(schema1.getKey(), userId, attrs1)
    attrRepo.addAttributes(schema2.getKey(), userId, attrs2)

    # 5. request Claims
    prover = Prover(ProverWalletInMemory(userId, publicRepo))
    claimsReqs = await prover.createClaimRequests([schemaId1, schemaId2])
    claims = await issuer.issueClaims(claimsReqs)
    await prover.processClaims(claims)

    # 6. proof Claims
    verifier = Verifier(WalletInMemory('verifier1', publicRepo))

    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid1': AttributeInfo('name', schema1.seqId)},
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('period', 5)})

    proof = await prover.presentProof(proofRequest)

    assert proof.requestedProof.revealed_attrs['attr_uuid1'][1] == 'Alex'
    assert await verifier.verify(proofRequest, proof)


@pytest.mark.asyncio
async def testSingleIssuerSingleProverPrimaryOnly(primes1):
    # 1. Init entities
    publicRepo = PublicRepoInMemory()
    attrRepo = AttributeRepoInMemory()
    issuer = Issuer(IssuerWalletInMemory('issuer1', publicRepo), attrRepo)

    # 2. Create a Schema
    schema = await issuer.genSchema('GVT', '1.0', GVT.attribNames())
    schemaId = ID(schema.getKey())

    # 3. Create keys for the Schema
    await issuer.genKeys(schemaId, **primes1)

    # 4. Issue accumulator
    await issuer.issueAccumulator(schemaId=schemaId, iA='110', L=5)

    # 4. set attributes for user1
    userId = '111'
    attrs = GVT.attribs(name='Alex', age=28, height=175, sex='male')
    attrRepo.addAttributes(schema.getKey(), userId, attrs)

    # 5. request Claims
    prover = Prover(ProverWalletInMemory(userId, publicRepo))
    claimsReq = await prover.createClaimRequest(schemaId, None, False)
    (claim_signature, claim_attributes) = await issuer.issueClaim(schemaId, claimsReq)
    await prover.processClaim(schemaId, claim_attributes, claim_signature)

    # 6. proof Claims
    verifier = Verifier(WalletInMemory('verifier1', publicRepo))

    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid1': AttributeInfo('name', schema.seqId)},
                                predicates={'predicate_uuid1': PredicateGE('age', 18)})

    proof = await prover.presentProof(proofRequest)

    assert proof.requestedProof.revealed_attrs['attr_uuid1'][1] == 'Alex'
    assert await verifier.verify(proofRequest, proof)
